#!/usr/bin/env python3

"""
Application settings utilities
"""

import json
import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE2

from cruiz.settings.managers.namedlocalcache import NamedLocalCacheSettingsReader

if PYSIDE2:
    from cruiz.pyside2.preferences import Ui_PreferencesDialog

    QAction = QtWidgets.QAction
else:
    from cruiz.pyside6.preferences import Ui_PreferencesDialog

    QAction = QtGui.QAction

from cruiz.settings.managers.generalpreferences import (
    GeneralSettings,
    GeneralSettingsReader,
    GeneralSettingsWriter,
)
from cruiz.settings.managers.fontpreferences import (
    FontUsage,
    FontSettings,
    FontSettingsReader,
    FontSettingsWriter,
)
from cruiz.settings.managers.conanpreferences import (
    ConanSettings,
    ConanSettingsReader,
    ConanSettingsWriter,
)
from cruiz.settings.managers.localcachepreferences import (
    LocalCacheSettings,
    LocalCacheSettingsReader,
    LocalCacheSettingsWriter,
)
from cruiz.settings.managers.namedlocalcache import (
    AllNamedLocalCacheSettingsReader,
    NamedLocalCacheDeleter,
)
from cruiz.settings.managers.graphvizpreferences import (
    GraphVizSettings,
    GraphVizSettingsReader,
    GraphVizSettingsWriter,
)
from cruiz.settings.managers.cmakepreferences import (
    CMakeSettings,
    CMakeSettingsReader,
    CMakeSettingsWriter,
)
from cruiz.settings.managers.ninjapreferences import (
    NinjaSettings,
    NinjaSettingsReader,
    NinjaSettingsWriter,
)
from cruiz.settings.managers.compilercachepreferences import (
    CompilerCacheSettings,
    CompilerCacheSettingsReader,
    CompilerCacheSettingsWriter,
)
from cruiz.settings.managers.recentrecipes import (
    RecentRecipeSettingsReader,
    RecentRecipeSettingsDeleter,
)
from cruiz.settings.managers.recipe import (
    RecipeSettingsReader,
    RecipeSettingsWriter,
    RecipeSettings,
    RecipeSettingsDeleter,
)
from cruiz.settings.managers.recentconanconfigs import (
    RecentConanConfigSettingsReader,
    RecentConanConfigSettingsDeleter,
)
from cruiz.settings.managers.recentconanremotes import (
    RecentConanRemotesSettingsReader,
    RecentConanRemotesSettingsDeleter,
)
from cruiz.settings.managers.shortcuts import (
    ShortcutSettings,
    ShortcutSettingsReader,
    ShortcutSettingsWriter,
)
from cruiz.settings.managers.cleansettings import sanitise_settings
from cruiz.settings.managers.restoredefaults import factory_reset

from cruiz.settings.models.recentconanremotesmodel import RecentConanRemotesModel
from cruiz.settings.models.recentconanconfigmodel import RecentConanConfigModel
from cruiz.settings.models.recipesmodel import RecipesModel

from cruiz.widgets.util import BlockSignals, search_for_file_options

from cruiz.constants import DEFAULT_CACHE_NAME


class PreferencesDialog(QtWidgets.QDialog):
    """
    Dialog representing the preferences.
    """

    modified = QtCore.Signal()
    preferences_updated = QtCore.Signal()

    def __init__(self) -> None:
        super().__init__()
        self._ui = Ui_PreferencesDialog()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        # -- signals
        self.modified.connect(self._modification_applied)
        self.preferences_updated.connect(
            qApp.on_preferences_updated  # type: ignore  # noqa: F821
        )
        # -- toolbox
        self._ui.prefs_toolbox.currentChanged.connect(self._current_load_defaults)
        self._import_prefs = QtWidgets.QPushButton("Import...")
        self._export_prefs = QtWidgets.QPushButton("Export...")
        self._clean_prefs = QtWidgets.QPushButton("Clean...")
        self._restore_default_prefs = QtWidgets.QPushButton("Restore defaults")
        self._setup_buttons()
        self._setup_general_toolbox()
        self._setup_font_toolbox()
        self._setup_conan_toolbox()
        self._setup_localcache_toolbox()
        self._setup_graphviz_toolbox()
        self._setup_cmake_toolbox()
        self._setup_compilercache_toolbox()
        self._setup_shortcuts_toolbox()
        self._setup_recipes_toolbox()
        self._setup_recentconfigs_toolbox()
        self._setup_recentremotes_toolbox()
        # -- load initial toolbox settings
        self._current_load_defaults(self._ui.prefs_toolbox.currentIndex())

    def _any_modifications(self) -> bool:
        return (
            not self._prefs_general.empty(GeneralSettingsReader())
            or not self._prefs_font[FontUsage.UI].empty(
                FontSettingsReader(FontUsage.UI)
            )
            or not self._prefs_font[FontUsage.OUTPUT].empty(
                FontSettingsReader(FontUsage.OUTPUT)
            )
            or not self._prefs_conan.empty(ConanSettingsReader())
            or not self._prefs_localcache.empty(LocalCacheSettingsReader())
            or not self._prefs_graphviz.empty(GraphVizSettingsReader())
            or not self._prefs_cmake.empty(CMakeSettingsReader())
            or not self._prefs_ninja.empty(NinjaSettingsReader())
            or not self._prefs_compilercache.empty(CompilerCacheSettingsReader())
            or not self._prefs_shortcuts.empty(ShortcutSettingsReader())
        )

    def _save(self) -> None:
        GeneralSettingsWriter().sync(self._prefs_general)
        FontSettingsWriter(FontUsage.UI).sync(self._prefs_font[FontUsage.UI])
        FontSettingsWriter(FontUsage.OUTPUT).sync(self._prefs_font[FontUsage.OUTPUT])
        ConanSettingsWriter().sync(self._prefs_conan)
        LocalCacheSettingsWriter().sync(self._prefs_localcache)
        GraphVizSettingsWriter().sync(self._prefs_graphviz)
        CMakeSettingsWriter().sync(self._prefs_cmake)
        NinjaSettingsWriter().sync(self._prefs_ninja)
        CompilerCacheSettingsWriter().sync(self._prefs_compilercache)
        ShortcutSettingsWriter().sync(self._prefs_shortcuts)
        self.modified.emit()
        self._current_load_defaults(self._ui.prefs_toolbox.currentIndex())
        self.preferences_updated.emit()

    def _modification_applied(self) -> None:
        enabled = self._any_modifications()
        self._ui.prefs_buttons.button(
            QtWidgets.QDialogButtonBox.StandardButton.Apply
        ).setEnabled(enabled)
        self._ui.prefs_buttons.button(
            QtWidgets.QDialogButtonBox.StandardButton.Ok
        ).setEnabled(enabled)

    def _setup_buttons(self) -> None:
        self._ui.prefs_buttons.button(
            QtWidgets.QDialogButtonBox.StandardButton.Apply
        ).clicked.connect(self._save)
        self._ui.prefs_buttons.button(
            QtWidgets.QDialogButtonBox.StandardButton.Ok
        ).clicked.connect(self._save)
        self._ui.prefs_buttons.button(
            QtWidgets.QDialogButtonBox.StandardButton.Ok
        ).clicked.connect(self.accept)
        self._ui.prefs_buttons.button(
            QtWidgets.QDialogButtonBox.StandardButton.Apply
        ).setEnabled(False)
        self._ui.prefs_buttons.button(
            QtWidgets.QDialogButtonBox.StandardButton.Ok
        ).setEnabled(False)
        self._ui.prefs_buttons.addButton(
            self._import_prefs, QtWidgets.QDialogButtonBox.ActionRole
        )
        self._ui.prefs_buttons.addButton(
            self._export_prefs, QtWidgets.QDialogButtonBox.ActionRole
        )
        self._ui.prefs_buttons.addButton(
            self._clean_prefs, QtWidgets.QDialogButtonBox.ActionRole
        )
        self._ui.prefs_buttons.addButton(
            self._restore_default_prefs, QtWidgets.QDialogButtonBox.ActionRole
        )
        self._import_prefs.clicked.connect(self._import_presets)
        self._export_prefs.clicked.connect(self._export_presets)
        self._clean_prefs.clicked.connect(self._clean_preferences)
        self._restore_default_prefs.clicked.connect(self._restore_defaults)

    def _setup_general_toolbox(self) -> None:
        self._prefs_general = GeneralSettings()
        self._ui.prefs_general_clearpanes.stateChanged.connect(
            self._general_clearplanes
        )
        self._ui.prefs_general_combine_panes.stateChanged.connect(
            self._general_combinepanes
        )
        self._ui.prefs_general_usebatching.stateChanged.connect(
            self._general_usebatching
        )
        self._ui.prefs_general_enable_wallclock.stateChanged.connect(
            self._general_wallclock
        )
        self._ui.prefs_general_enable_darkmode.stateChanged.connect(
            self._general_darkmode
        )
        self._ui.prefs_general_enable_compact.stateChanged.connect(
            self._general_compactlook
        )
        dir_icon = self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon)
        open_recipedir_action = QAction(dir_icon, "", self)
        open_recipedir_action.triggered.connect(self._general_open_recipedir)
        self._ui.prefs_general_default_recipe_dir.addAction(
            open_recipedir_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        self._ui.prefs_general_default_recipe_dir.textChanged.connect(
            self._general_recipedir
        )
        self._ui.prefs_general_busy_colour.clicked.connect(self._general_busyiconcolour)
        self._ui.prefs_general_found_text_background_colour.clicked.connect(
            self._general_foundtextbgroundcolour
        )
        open_recipe_in_editor_action = QAction(dir_icon, "", self)
        open_recipe_in_editor_action.triggered.connect(self._general_open_recipe_editor)
        self._ui.prefs_general_recipe_editor.addAction(
            open_recipe_in_editor_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        self._ui.prefs_general_recipe_editor.textChanged.connect(
            self._general_recipe_editor
        )
        self._ui.prefs_general_new_recipe_load.stateChanged.connect(
            self._general_newrecipeload
        )  # TODO: this has no place in the new UI

    def _setup_font_toolbox(self) -> None:
        self._prefs_font = {
            FontUsage.UI: FontSettings(),
            FontUsage.OUTPUT: FontSettings(),
        }
        self._ui.prefs_font_ui_change.clicked.connect(self._font_change_ui)
        self._ui.prefs_font_ui_reset.clicked.connect(self._font_reset_ui)
        self._ui.prefs_font_output_change.clicked.connect(self._font_change_output)
        self._ui.prefs_font_output_reset.clicked.connect(self._font_reset_output)

    def _setup_conan_toolbox(self) -> None:
        self._prefs_conan = ConanSettings()
        self._ui.prefs_conan_log_level.currentTextChanged.connect(
            self._conan_change_log_level
        )
        self._ui.prefs_conan_version_list_path_segment.textChanged.connect(
            self._conan_version_list_path_segment_changed
        )

    def _setup_localcache_toolbox(self) -> None:
        self._prefs_localcache = LocalCacheSettings()
        self._ui.prefs_localcache_config_to_install.textChanged.connect(
            self._localcache_install_path_changed
        )
        self._ui.prefs_localcache_config_git_branch.textChanged.connect(
            self._localcache_install_branch_changed
        )
        self._ui.prefs_localcache_forget_cache.currentTextChanged.connect(
            self._localcache_change_forget_cache
        )
        self._ui.prefs_localcache_do_forget.clicked.connect(
            self._localcache_forget_cache
        )

    def _setup_graphviz_toolbox(self) -> None:
        self._prefs_graphviz = GraphVizSettings()
        self._ui.prefs_graphviz_bin_directory.textChanged.connect(
            self._graphviz_bin_directory_changed
        )
        open_graphviz_bindir_action = QAction(
            self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon), "", self
        )
        open_graphviz_bindir_action.triggered.connect(self._graphviz_open_bindir)
        self._ui.prefs_graphviz_bin_directory.addAction(
            open_graphviz_bindir_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )

    def _setup_cmake_toolbox(self) -> None:
        self._prefs_cmake = CMakeSettings()
        self._prefs_ninja = NinjaSettings()
        self._ui.prefs_cmake_cmake_bin_directory.textChanged.connect(
            self._cmake_bin_directory_changed
        )
        dir_icon = self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon)
        open_cmake_bindir_action = QAction(dir_icon, "", self)
        open_cmake_bindir_action.triggered.connect(self._cmake_open_bindir)
        self._ui.prefs_cmake_cmake_bin_directory.addAction(
            open_cmake_bindir_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        self._ui.prefs_cmake_ninja_bin_directory.textChanged.connect(
            self._ninja_bin_directory_changed
        )
        open_ninja_bindir_action = QAction(dir_icon, "", self)
        open_ninja_bindir_action.triggered.connect(self._ninja_open_bindir)
        self._ui.prefs_cmake_ninja_bin_directory.addAction(
            open_ninja_bindir_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )

    def _setup_compilercache_toolbox(self) -> None:
        self._prefs_compilercache = CompilerCacheSettings()
        self._ui.prefs_compilercache_default.currentTextChanged.connect(
            self._compilercache_default_changed
        )
        self._ui.prefs_compilercache_ccache_location.textChanged.connect(
            self._compilercache_ccache_bin_directory_changed
        )
        dir_icon = self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon)
        open_ccache_bindir_action = QAction(dir_icon, "", self)
        open_ccache_bindir_action.triggered.connect(
            self._compilercache_open_ccache_bindir
        )
        self._ui.prefs_compilercache_ccache_location.addAction(
            open_ccache_bindir_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        self._ui.prefs_compilercache_sccache_location.textChanged.connect(
            self._compilercache_sccache_bin_directory_changed
        )
        open_scache_bindir_action = QAction(dir_icon, "", self)
        open_scache_bindir_action.triggered.connect(
            self._compilercache_open_sccache_bindir
        )
        self._ui.prefs_compilercache_sccache_location.addAction(
            open_scache_bindir_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        self._ui.prefs_compilercache_buildcache_location.textChanged.connect(
            self._compilercache_buildcache_bin_directory_changed
        )
        open_buildcache_bindir_action = QAction(dir_icon, "", self)
        open_buildcache_bindir_action.triggered.connect(
            self._compilercache_open_buildcache_bindir
        )
        self._ui.prefs_compilercache_buildcache_location.addAction(
            open_buildcache_bindir_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )

    def _setup_shortcuts_toolbox(self) -> None:
        self._prefs_shortcuts = ShortcutSettings()
        self._ui.prefs_shortcuts_create_edit.textChanged.connect(
            self._shortcuts_change_create
        )
        self._ui.prefs_shortcuts_imports_edit.textChanged.connect(
            self._shortcuts_change_imports
        )
        self._ui.prefs_shortcuts_install_edit.textChanged.connect(
            self._shortcuts_change_install
        )
        self._ui.prefs_shortcuts_installupdates_edit.textChanged.connect(
            self._shortcuts_change_installupdates
        )
        self._ui.prefs_shortcuts_source_edit.textChanged.connect(
            self._shortcuts_change_source
        )
        self._ui.prefs_shortcuts_build_edit.textChanged.connect(
            self._shortcuts_change_build
        )
        self._ui.prefs_shortcuts_package_edit.textChanged.connect(
            self._shortcuts_change_package
        )
        self._ui.prefs_shortcuts_exportpackage_edit.textChanged.connect(
            self._shortcuts_change_exportpkg
        )
        self._ui.prefs_shortcuts_test_edit.textChanged.connect(
            self._shortcuts_change_test
        )
        self._ui.prefs_shortcuts_remove_edit.textChanged.connect(
            self._shortcuts_change_remove
        )
        self._ui.prefs_shortcuts_cancel_edit.textChanged.connect(
            self._shortcuts_change_cancel
        )
        self._ui.prefs_shortcuts_cmakebuildtool_edit.textChanged.connect(
            self._shortcuts_change_cmakebuildtool
        )
        self._ui.prefs_shortcuts_cmakebuildtoolverbose_edit.textChanged.connect(
            self._shortcuts_change_cmakebuildtoolverbose
        )
        self._ui.prefs_shortcuts_deletecmakecache_edit.textChanged.connect(
            self._shortcuts_change_deletecmakecache
        )

    def _setup_recipes_toolbox(self) -> None:
        self._prefs_recipes_model = RecipesModel()
        self._ui.prefs_recipes_table.setModel(self._prefs_recipes_model)
        self._ui.prefs_recipes_table.horizontalHeader().setSectionResizeMode(
            0, QtWidgets.QHeaderView.ResizeToContents
        )
        self._ui.prefs_recipes_table.horizontalHeader().setSectionResizeMode(
            1, QtWidgets.QHeaderView.ResizeToContents
        )
        self._ui.prefs_recipes_table.addAction(self._ui.actionForget_recipe)
        self._ui.prefs_recipes_table.customContextMenuRequested.connect(
            self._recipes_context_menu
        )

    def _setup_recentconfigs_toolbox(self) -> None:
        self._prefs_recentconfigs_model = RecentConanConfigModel()
        self._ui.prefs_recentconfigs_list.setModel(self._prefs_recentconfigs_model)
        self._ui.prefs_recentconfigs_list.addAction(self._ui.actionForget_config)
        self._ui.prefs_recentconfigs_list.customContextMenuRequested.connect(
            self._recentconfigs_context_menu
        )

    def _setup_recentremotes_toolbox(self) -> None:
        self._prefs_recentremotes_model = RecentConanRemotesModel()
        self._ui.prefs_recentremotes_list.setModel(self._prefs_recentremotes_model)
        self._ui.prefs_recentremotes_list.addAction(self._ui.actionForget_remote)
        self._ui.prefs_recentremotes_list.customContextMenuRequested.connect(
            self._recentremotes_context_menu
        )

    def _current_load_defaults(self, index: int) -> None:
        widget = self._ui.prefs_toolbox.widget(index)
        name = widget.objectName()
        if name == "prefs_general":
            self._general_load_defaults()
        elif name == "prefs_fonts":
            self._font_load_defaults()
        elif name == "prefs_conan":
            self._conan_load_defaults()
        elif name == "prefs_localcache":
            self._localcache_load_defaults()
        elif name == "prefs_graphviz":
            self._graphviz_load_defaults()
        elif name == "prefs_cmake":
            self._cmake_load_defaults()
        elif name == "prefs_compilercache":
            self._compilercache_load_defaults()
        elif name == "prefs_shortcuts":
            self._shortcuts_load_defaults()
        elif name == "prefs_recipes":
            self._recipes_load_defaults()
        elif name == "prefs_recentconfigs":
            self._recentconfigs_load_defaults()
        elif name == "prefs_recentremotes":
            self._recentremotes_load_defaults()
        else:
            raise RuntimeError(f"Unknown toolbox widget with name '{name}")

    def reject(self) -> None:
        if self._any_modifications():
            response = QtWidgets.QMessageBox.question(
                self,
                "Unsaved preferences",
                "Modifications are unsaved. Do you want to discard them?",
            )
            if response == QtWidgets.QMessageBox.No:
                return
        super().reject()

    # -- general --

    def _update_busy_icon_colour(self, colour: QtGui.QColor) -> None:
        self._ui.prefs_general_busy_colour.setStyleSheet(
            "background-color: " f"rgb({colour.red()},{colour.green()},{colour.blue()})"
        )

    def _update_found_text_background_colour(self, colour: QtGui.QColor) -> None:
        self._ui.prefs_general_found_text_background_colour.setStyleSheet(
            "background-color: " f"rgb({colour.red()},{colour.green()},{colour.blue()})"
        )

    def _general_load_defaults(self) -> None:
        with GeneralSettingsReader() as settings:
            with BlockSignals(self._ui.prefs_general_clearpanes) as blocked_widget:
                blocked_widget.setChecked(settings.clear_panes.resolve())
            with BlockSignals(self._ui.prefs_general_combine_panes) as blocked_widget:
                blocked_widget.setChecked(settings.combine_panes.resolve())
            with BlockSignals(self._ui.prefs_general_usebatching) as blocked_widget:
                blocked_widget.setChecked(settings.use_stdout_batching.resolve())
            with BlockSignals(
                self._ui.prefs_general_enable_wallclock
            ) as blocked_widget:
                blocked_widget.setChecked(settings.enable_command_timing.resolve())
            with BlockSignals(self._ui.prefs_general_enable_darkmode) as blocked_widget:
                blocked_widget.setChecked(settings.use_dark_mode.resolve())
            with BlockSignals(self._ui.prefs_general_enable_compact) as blocked_widget:
                blocked_widget.setChecked(settings.use_compact_look.resolve())
            with BlockSignals(
                self._ui.prefs_general_default_recipe_dir
            ) as blocked_widget:
                blocked_widget.setText(
                    settings.default_recipe_directory.resolve() or ""
                )
            with BlockSignals(self._ui.prefs_general_busy_colour):
                colour = settings.busy_icon_colour.resolve()
                self._update_busy_icon_colour(colour)
            with BlockSignals(self._ui.prefs_general_found_text_background_colour):
                colour = settings.found_text_background_colour.resolve()
                self._update_found_text_background_colour(colour)
            with BlockSignals(self._ui.prefs_general_recipe_editor) as blocked_widget:
                blocked_widget.setText(settings.default_recipe_editor.resolve() or "")
            # Note: the following is not part of the new UI
            with BlockSignals(self._ui.prefs_general_new_recipe_load) as blocked_widget:
                blocked_widget.setChecked(
                    settings.new_recipe_loading_behaviour.resolve()
                )

    def _general_change_boolean(self, property_name: str, state: int) -> None:
        enabled = state == QtCore.Qt.Checked  # type: ignore[comparison-overlap]
        setattr(self._prefs_general, property_name, enabled)
        self.modified.emit()

    def _general_clearplanes(self, state: int) -> None:
        self._prefs_general.clear_panes = state == QtCore.Qt.Checked  # type: ignore
        self.modified.emit()

    def _general_combinepanes(self, state: int) -> None:
        self._prefs_general.combine_panes = state == QtCore.Qt.Checked  # type: ignore
        self.modified.emit()
        self._requires_restart()

    def _general_usebatching(self, state: int) -> None:
        enabled = state == QtCore.Qt.Checked  # type: ignore[comparison-overlap]
        self._prefs_general.use_stdout_batching = enabled  # type: ignore
        self.modified.emit()
        self._requires_restart()

    def _general_wallclock(self, state: int) -> None:
        enabled = state == QtCore.Qt.Checked  # type: ignore[comparison-overlap]
        self._prefs_general.enable_command_timing = enabled  # type: ignore
        self.modified.emit()

    def _general_darkmode(self, state: int) -> None:
        self._prefs_general.use_dark_mode = state == QtCore.Qt.Checked  # type: ignore
        self.modified.emit()

    def _general_compactlook(self, state: int) -> None:
        enabled = state == QtCore.Qt.Checked  # type: ignore[comparison-overlap]
        self._prefs_general.use_compact_look = enabled  # type: ignore
        self.modified.emit()

    def _general_open_recipedir(self) -> None:
        with GeneralSettingsReader() as settings:
            last_dir = settings.default_recipe_directory.resolve()
        new_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            caption="Select default recipe directory",
            dir=last_dir,
        )
        if new_dir:
            self._ui.prefs_general_default_recipe_dir.setText(new_dir)

    def _general_recipedir(self, text: str) -> None:
        self._prefs_general.default_recipe_directory = text or None  # type: ignore
        self.modified.emit()

    def _general_busyiconcolour(self) -> None:
        with GeneralSettingsReader() as settings:
            busy_color = settings.busy_icon_colour.resolve()
        dialog = QtWidgets.QColorDialog(busy_color, self)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            new_color = dialog.currentColor()
            self._update_busy_icon_colour(new_color)
            self._prefs_general.busy_icon_colour = new_color  # type: ignore[assignment]
            self.modified.emit()

    def _general_foundtextbgroundcolour(self) -> None:
        with GeneralSettingsReader() as settings:
            color = settings.found_text_background_colour.resolve()
        dialog = QtWidgets.QColorDialog(color, self)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            new_color = dialog.currentColor()
            self._update_found_text_background_colour(new_color)
            self._prefs_general.found_text_background_colour = (
                new_color  # type: ignore[assignment]
            )
            self.modified.emit()

    def _general_open_recipe_editor(self) -> None:
        with GeneralSettingsReader() as settings:
            last_dir = settings.default_recipe_editor.resolve()
        new_path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            caption="Select default recipe editor",
            dir=last_dir,
            options=search_for_file_options(),
        )
        if new_path:
            self._ui.prefs_general_recipe_editor.setText(new_path)

    def _general_recipe_editor(self, text: str) -> None:
        self._prefs_general.default_recipe_editor = text or None  # type: ignore
        self.modified.emit()

    def _general_newrecipeload(self, state: int) -> None:
        enabled = state == QtCore.Qt.Checked  # type: ignore[comparison-overlap]
        self._prefs_general.new_recipe_loading_behaviour = enabled  # type: ignore
        self.modified.emit()

    # -- font --
    @staticmethod
    def _font_from_details(
        details: typing.Tuple[typing.Optional[str], typing.Optional[int]]
    ) -> typing.Optional[QtGui.QFont]:
        font = QtGui.QFont(*details) if details[0] else None  # type: ignore
        return font

    @staticmethod
    def _font_description(font: QtGui.QFont) -> str:
        return f"{font.family()} at size {font.pointSize()}"

    def _font_set_ui_preview(self, font: QtGui.QFont) -> None:
        self._ui.prefs_font_ui_label.setText(PreferencesDialog._font_description(font))
        self._ui.prefs_font_ui_preview.setFont(font)

    def _font_set_output_preview(self, font: QtGui.QFont) -> None:
        self._ui.prefs_font_output_label.setText(
            PreferencesDialog._font_description(font)
        )
        self._ui.prefs_font_output_preview.setFont(font)

    def _font_load_defaults(self) -> None:
        with FontSettingsReader(FontUsage.UI) as settings:
            font = (
                PreferencesDialog._font_from_details(
                    (settings.name.resolve(), settings.size.resolve())
                )
                or qApp.default_font  # type: ignore  # noqa: F821
            )
        self._font_set_ui_preview(font)

        with FontSettingsReader(FontUsage.OUTPUT) as settings:
            font = PreferencesDialog._font_from_details(
                (settings.name.resolve(), settings.size.resolve())
            ) or QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont)
        self._font_set_output_preview(font)

    def _font_change(self, usage: FontUsage, default_font: QtGui.QFont) -> None:
        with FontSettingsReader(usage) as settings:
            desired_font = (
                PreferencesDialog._font_from_details(
                    (settings.name.resolve(), settings.size.resolve())
                )
                or default_font
            )
        is_ok, font = (
            QtWidgets.QFontDialog.getFont(desired_font, self)
            if desired_font
            else QtWidgets.QFontDialog.getFont(self)
        )
        if is_ok:
            """
            with FontWriteAccess(usage) as fontsettings:
                fontsettings.name = font.family()
                fontsettings.size = font.pointSize()
                self._font_load_defaults()
            """
            assert isinstance(font, QtGui.QFont)
            self._prefs_font[usage].name = font.family()  # type: ignore
            self._prefs_font[usage].size = font.pointSize()  # type: ignore
            if usage == FontUsage.UI:
                self._font_set_ui_preview(font)
            elif usage == FontUsage.OUTPUT:
                self._font_set_output_preview(font)
            self.modified.emit()

    def _font_change_ui(self) -> None:
        self._font_change(FontUsage.UI, qApp.default_font)  # type: ignore # noqa: F821

    def _font_change_output(self) -> None:
        self._font_change(
            FontUsage.OUTPUT,
            QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont),
        )

    def _font_reset(self, usage: FontUsage) -> None:
        """
        with FontWriteAccess(usage) as fontsettings:
            del fontsettings.name
            del fontsettings.size
            self._font_load_defaults()
        """
        self._prefs_font[usage].name = None  # type: ignore
        self._prefs_font[usage].size = None  # type: ignore
        if usage == FontUsage.UI:
            self._font_set_ui_preview(qApp.default_font)  # type: ignore # noqa: F821
        elif usage == FontUsage.OUTPUT:
            self._font_set_output_preview(
                QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont)
            )
        self.modified.emit()

    def _font_reset_ui(self) -> None:
        self._font_reset(FontUsage.UI)

    def _font_reset_output(self) -> None:
        self._font_reset(FontUsage.OUTPUT)

    # -- conan --
    def _conan_load_defaults(self) -> None:
        with ConanSettingsReader() as settings:
            log_level = settings.log_level.resolve()
            conandata_version_yaml_pathsegment = (
                settings.conandata_version_yaml_pathsegment.resolve()
            )
        with BlockSignals(self._ui.prefs_conan_log_level) as blocked_widget:
            blocked_widget.setCurrentText(log_level)
        with BlockSignals(
            self._ui.prefs_conan_version_list_path_segment
        ) as blocked_widget:
            blocked_widget.setText(conandata_version_yaml_pathsegment)

    def _conan_change_log_level(self, text: str) -> None:
        self._prefs_conan.log_level = text  # type: ignore
        self.modified.emit()

    def _conan_version_list_path_segment_changed(self, text: str) -> None:
        value = text or None
        self._prefs_conan.conandata_version_yaml_pathsegment = value  # type: ignore
        self.modified.emit()

    # -- local cache --
    def _localcache_load_defaults(self) -> None:
        with LocalCacheSettingsReader() as settings:
            install_path = settings.new_configuration_install.resolve()
            git_branch = settings.new_configuration_git_branch.resolve()
        with BlockSignals(
            self._ui.prefs_localcache_config_to_install
        ) as blocked_widget:
            blocked_widget.setText(install_path)
        with BlockSignals(
            self._ui.prefs_localcache_config_git_branch
        ) as blocked_widget:
            blocked_widget.setText(git_branch)
        with BlockSignals(self._ui.prefs_localcache_forget_cache) as blocked_widget:
            blocked_widget.clear()
            with AllNamedLocalCacheSettingsReader() as names:
                blocked_widget.addItems(names)
        self._localcache_change_forget_cache(
            self._ui.prefs_localcache_forget_cache.currentText()
        )

    def _localcache_install_path_changed(self, text: str) -> None:
        self._prefs_localcache.new_configuration_install = text or None  # type: ignore
        self.modified.emit()

    def _localcache_install_branch_changed(self, text: str) -> None:
        value = text or None
        self._prefs_localcache.new_configuration_git_branch = value  # type: ignore
        self.modified.emit()

    def _localcache_change_forget_cache(self, text: str) -> None:
        with NamedLocalCacheSettingsReader(text) as settings:
            uuids = settings.recipe_uuids
        self._ui.prefs_localcache_do_forget.setEnabled(
            not uuids and text != DEFAULT_CACHE_NAME
        )

    def _localcache_forget_cache(self) -> None:
        NamedLocalCacheDeleter().delete(
            self._ui.prefs_localcache_forget_cache.currentText()
        )

    # -- graphviz --
    def _graphviz_load_defaults(self) -> None:
        with GraphVizSettingsReader() as settings:
            bin_path = settings.bin_directory.resolve()
        with BlockSignals(self._ui.prefs_graphviz_bin_directory) as blocked_widget:
            blocked_widget.setText(bin_path)

    def _graphviz_bin_directory_changed(self, text: str) -> None:
        self._prefs_graphviz.bin_directory = text or None  # type: ignore
        self.modified.emit()

    def _graphviz_open_bindir(self) -> None:
        with GraphVizSettingsReader() as settings:
            last_dir = settings.bin_directory.resolve()
        new_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            caption="Select GraphViz bin directory",
            dir=last_dir,
        )
        if new_dir:
            self._ui.prefs_graphviz_bin_directory.setText(new_dir)

    # -- CMake & Ninja --
    def _cmake_load_defaults(self) -> None:
        with CMakeSettingsReader() as settings:
            cmake_bin_path = settings.bin_directory.resolve()
        with BlockSignals(self._ui.prefs_cmake_cmake_bin_directory) as blocked_widget:
            blocked_widget.setText(cmake_bin_path)
        with NinjaSettingsReader() as settings:
            ninja_bin_path = settings.bin_directory.resolve()
        with BlockSignals(self._ui.prefs_cmake_ninja_bin_directory) as blocked_widget:
            blocked_widget.setText(ninja_bin_path)

    def _cmake_bin_directory_changed(self, text: str) -> None:
        self._prefs_cmake.bin_directory = text or None  # type: ignore
        self.modified.emit()

    def _cmake_open_bindir(self) -> None:
        with CMakeSettingsReader() as settings:
            last_dir = settings.bin_directory.resolve()
        new_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            caption="Select CMake bin directory",
            dir=last_dir,
        )
        if new_dir:
            self._ui.prefs_cmake_cmake_bin_directory.setText(new_dir)

    def _ninja_bin_directory_changed(self, text: str) -> None:
        self._prefs_ninja.bin_directory = text or None  # type: ignore
        self.modified.emit()

    def _ninja_open_bindir(self) -> None:
        with NinjaSettingsReader() as settings:
            last_dir = settings.bin_directory.resolve()
        new_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            caption="Select Ninja bin directory",
            dir=last_dir,
        )
        if new_dir:
            self._ui.prefs_cmake_ninja_bin_directory.setText(new_dir)

    # -- Compiler cache --
    def _compilercache_load_defaults(self) -> None:
        with CompilerCacheSettingsReader() as settings:
            compilercache_default = settings.default.resolve()
            ccache_bin_dir = settings.ccache_bin_directory.resolve()
            sccache_bin_dir = settings.sccache_bin_directory.resolve()
            buildcache_bin_dir = settings.buildcache_bin_directory.resolve()
        with BlockSignals(self._ui.prefs_compilercache_default) as blocked_widget:
            blocked_widget.setCurrentText(compilercache_default)
        with BlockSignals(
            self._ui.prefs_compilercache_ccache_location
        ) as blocked_widget:
            blocked_widget.setText(ccache_bin_dir)
        with BlockSignals(
            self._ui.prefs_compilercache_sccache_location
        ) as blocked_widget:
            blocked_widget.setText(sccache_bin_dir)
        with BlockSignals(
            self._ui.prefs_compilercache_buildcache_location
        ) as blocked_widget:
            blocked_widget.setText(buildcache_bin_dir)

    def _compilercache_default_changed(self, text: str) -> None:
        self._prefs_compilercache.default = text  # type: ignore
        self.modified.emit()

    def _compilercache_ccache_bin_directory_changed(self, text: str) -> None:
        self._prefs_compilercache.ccache_bin_directory = text or None  # type: ignore
        self.modified.emit()

    def _compilercache_open_ccache_bindir(self) -> None:
        with CompilerCacheSettingsReader() as settings:
            last_dir = settings.ccache_bin_directory.resolve()
        new_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            caption="Select ccache bin directory",
            dir=last_dir,
        )
        if new_dir:
            self._ui.prefs_compilercache_ccache_location.setText(new_dir)

    def _compilercache_sccache_bin_directory_changed(self, text: str) -> None:
        self._prefs_compilercache.sccache_bin_directory = text or None  # type: ignore
        self.modified.emit()

    def _compilercache_open_sccache_bindir(self) -> None:
        with CompilerCacheSettingsReader() as settings:
            last_dir = settings.sccache_bin_directory.resolve()
        new_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            caption="Select sccache bin directory",
            dir=last_dir,
        )
        if new_dir:
            self._ui.prefs_compilercache_sccache_location.setText(new_dir)

    def _compilercache_buildcache_bin_directory_changed(self, text: str) -> None:
        value = text or None
        self._prefs_compilercache.buildcache_bin_directory = value  # type: ignore
        self.modified.emit()

    def _compilercache_open_buildcache_bindir(self) -> None:
        with CompilerCacheSettingsReader() as settings:
            last_dir = settings.buildcache_bin_directory.resolve()
        new_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            caption="Select buildcache bin directory",
            dir=last_dir,
        )
        if new_dir:
            self._ui.prefs_compilercache_buildcache_location.setText(new_dir)

    # -- Shortcuts --
    def _shortcuts_load_defaults(self) -> None:
        with ShortcutSettingsReader() as settings:
            conan_create = settings.conan_create
            conan_create_updates = settings.conan_create_updates
            conan_imports = settings.conan_imports
            conan_install = settings.conan_install
            conan_install_updates = settings.conan_install_updates
            conan_source = settings.conan_source
            conan_build = settings.conan_build
            conan_package = settings.conan_package
            conan_exportpkg = settings.conan_export_package
            conan_test = settings.conan_test_package
            conan_remove = settings.conan_remove_package
            cancel = settings.cancel
            cmake_build_tool = settings.cmake_build_tool
            cmake_build_tool_verbose = settings.cmake_build_tool_verbose
            remove_cmakecache = settings.delete_cmake_cache

        with BlockSignals(self._ui.prefs_shortcuts_create_edit) as blocked_widget:
            blocked_widget.setPlaceholderText(conan_create.fallback)
            blocked_widget.setText(conan_create.value or "")
        with BlockSignals(
            self._ui.prefs_shortcuts_createupdates_edit
        ) as blocked_widget:
            blocked_widget.setPlaceholderText(conan_create_updates.fallback)
            blocked_widget.setText(conan_create_updates.value or "")
        with BlockSignals(self._ui.prefs_shortcuts_imports_edit) as blocked_widget:
            blocked_widget.setPlaceholderText(conan_imports.fallback)
            blocked_widget.setText(conan_imports.value or "")
        with BlockSignals(self._ui.prefs_shortcuts_install_edit) as blocked_widget:
            blocked_widget.setPlaceholderText(conan_install.fallback)
            blocked_widget.setText(conan_install.value or "")
        with BlockSignals(
            self._ui.prefs_shortcuts_installupdates_edit
        ) as blocked_widget:
            blocked_widget.setPlaceholderText(conan_install_updates.fallback)
            blocked_widget.setText(conan_install_updates.value or "")
        with BlockSignals(self._ui.prefs_shortcuts_source_edit) as blocked_widget:
            blocked_widget.setPlaceholderText(conan_source.fallback)
            blocked_widget.setText(conan_source.value or "")
        with BlockSignals(self._ui.prefs_shortcuts_build_edit) as blocked_widget:
            blocked_widget.setPlaceholderText(conan_build.fallback)
            blocked_widget.setText(conan_build.value or "")
        with BlockSignals(self._ui.prefs_shortcuts_package_edit) as blocked_widget:
            blocked_widget.setPlaceholderText(conan_package.fallback)
            blocked_widget.setText(conan_package.value or "")
        with BlockSignals(
            self._ui.prefs_shortcuts_exportpackage_edit
        ) as blocked_widget:
            blocked_widget.setPlaceholderText(conan_exportpkg.fallback)
            blocked_widget.setText(conan_exportpkg.value or "")
        with BlockSignals(self._ui.prefs_shortcuts_test_edit) as blocked_widget:
            blocked_widget.setPlaceholderText(conan_test.fallback)
            blocked_widget.setText(conan_test.value or "")
        with BlockSignals(self._ui.prefs_shortcuts_remove_edit) as blocked_widget:
            blocked_widget.setPlaceholderText(conan_remove.fallback)
            blocked_widget.setText(conan_remove.value or "")
        with BlockSignals(self._ui.prefs_shortcuts_cancel_edit) as blocked_widget:
            blocked_widget.setPlaceholderText(cancel.fallback)
            blocked_widget.setText(cancel.value or "")
        with BlockSignals(
            self._ui.prefs_shortcuts_cmakebuildtool_edit
        ) as blocked_widget:
            blocked_widget.setPlaceholderText(cmake_build_tool.fallback)
            blocked_widget.setText(cmake_build_tool.value or "")
        with BlockSignals(
            self._ui.prefs_shortcuts_cmakebuildtoolverbose_edit
        ) as blocked_widget:
            blocked_widget.setPlaceholderText(cmake_build_tool_verbose.fallback)
            blocked_widget.setText(cmake_build_tool_verbose.value or "")
        with BlockSignals(
            self._ui.prefs_shortcuts_deletecmakecache_edit
        ) as blocked_widget:
            blocked_widget.setPlaceholderText(remove_cmakecache.fallback)
            blocked_widget.setText(remove_cmakecache.value or "")

    def _shortcuts_change_create(self, text: str) -> None:
        self._prefs_shortcuts.conan_create = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_imports(self, text: str) -> None:
        self._prefs_shortcuts.conan_imports = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_install(self, text: str) -> None:
        self._prefs_shortcuts.conan_install = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_installupdates(self, text: str) -> None:
        self._prefs_shortcuts.conan_install_updates = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_source(self, text: str) -> None:
        self._prefs_shortcuts.conan_source = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_build(self, text: str) -> None:
        self._prefs_shortcuts.conan_build = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_package(self, text: str) -> None:
        self._prefs_shortcuts.conan_package = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_exportpkg(self, text: str) -> None:
        self._prefs_shortcuts.conan_export_package = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_test(self, text: str) -> None:
        self._prefs_shortcuts.conan_test_package = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_remove(self, text: str) -> None:
        self._prefs_shortcuts.conan_remove_package = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_cancel(self, text: str) -> None:
        self._prefs_shortcuts.cancel = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_cmakebuildtool(self, text: str) -> None:
        self._prefs_shortcuts.cmake_build_tool = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_cmakebuildtoolverbose(self, text: str) -> None:
        self._prefs_shortcuts.cmake_build_tool_verbose = text or None  # type: ignore
        self.modified.emit()

    def _shortcuts_change_deletecmakecache(self, text: str) -> None:
        self._prefs_shortcuts.delete_cmake_cache = text or None  # type: ignore
        self.modified.emit()

    # -- Recipes --
    def _recipes_load_defaults(self) -> None:
        with RecentRecipeSettingsReader() as settings:
            self._prefs_recipes_model.set_uuids(settings.uuids())

    def _recipes_context_menu(self, position: QtCore.QPoint) -> None:
        selected_rows = self._ui.prefs_recipes_table.selectionModel().selectedRows()
        if not selected_rows:
            return
        selected_index = selected_rows[0]
        uuid = self._prefs_recipes_model.uuid(selected_index.row())
        menu = QtWidgets.QMenu(self)
        menu.addActions(self._ui.prefs_recipes_table.actions())
        forget_recipe_action = menu.actions()[0]
        forget_recipe_action.setData(uuid)
        forget_recipe_action.triggered.connect(self._recipes_forget_recipe)
        menu.addSeparator()
        menu.addMenu(self._recipes_build_change_cache_menu(uuid)).setText(
            "Change local cache to"
        )
        menu.exec_(self._ui.prefs_recipes_table.viewport().mapToGlobal(position))

    def _recipes_build_change_cache_menu(self, uuid: QtCore.QUuid) -> QtWidgets.QMenu:
        menu = QtWidgets.QMenu(self)
        with RecipeSettingsReader.from_uuid(uuid) as settings:
            uuid_local_cache = settings.local_cache_name.resolve()
        with AllNamedLocalCacheSettingsReader() as cache_names:
            for name in cache_names:
                if name == uuid_local_cache:
                    continue
                action = QAction(name, self)
                action.setData(uuid)
                action.triggered.connect(self._recipes_change_cache)
                menu.addAction(action)
        return menu

    def _recipes_change_cache(self) -> None:
        selected_rows = self._ui.prefs_recipes_table.selectionModel().selectedRows()
        if not selected_rows:
            return
        selected_index = selected_rows[0]
        path_itemdata = self._prefs_recipes_model.itemData(selected_index)
        if not path_itemdata:
            return
        cache_menu_action = self.sender()
        result = QtWidgets.QMessageBox.question(
            self,
            "Change local cache for recipe",
            "Are you sure you want to change the local cache for recipe "
            f"'{path_itemdata[0]}' to '{cache_menu_action.text()}'",
        )
        if result == QtWidgets.QMessageBox.StandardButton.No:
            return
        settings = RecipeSettings()
        settings.local_cache_name = cache_menu_action.text()
        RecipeSettingsWriter.from_uuid(cache_menu_action.data()).sync(settings)
        self._recipes_load_defaults()

    def _recipes_forget_recipe(self) -> None:
        action = self.sender()
        uuid = action.data()
        with RecipeSettingsReader.from_uuid(uuid) as settings:
            has_editables = settings.editables_count.resolve()
            recipe_path = settings.path.resolve()
        if has_editables:
            QtWidgets.QMessageBox.critical(
                self,
                "Recipe still has editables",
                f"Recipe {recipe_path} still has editable dependencies. "
                "Cannot forget until they are removed.",
            )
            return
        # TODO: this is not ideal, as it's not atomic
        # there are two QSettings created and destroyed
        # so there is a possibility that one part can fail,
        # so the settings can be left in a inconsistent state
        RecentRecipeSettingsDeleter().delete(uuid)
        RecipeSettingsDeleter().delete(uuid)
        self._recipes_load_defaults()

    # -- Recent configs --
    def _recentconfigs_load_defaults(self) -> None:
        with RecentConanConfigSettingsReader() as settings:
            self._prefs_recentconfigs_model.set_paths(settings.paths.resolve())

    def _recentconfigs_context_menu(self, position: QtCore.QPoint) -> None:
        selected_rows = (
            self._ui.prefs_recentconfigs_list.selectionModel().selectedRows()
        )
        if not selected_rows:
            return
        selected_index = selected_rows[0]
        path = self._prefs_recentconfigs_model.path(selected_index.row())
        menu = QtWidgets.QMenu(self)
        menu.addActions(self._ui.prefs_recentconfigs_list.actions())
        forget_config_action = menu.actions()[0]
        forget_config_action.setData(path)
        forget_config_action.triggered.connect(self._recentconfigs_forget_config)
        menu.exec_(self._ui.prefs_recentconfigs_list.viewport().mapToGlobal(position))

    def _recentconfigs_forget_config(self) -> None:
        action = self.sender()
        path = action.data()
        RecentConanConfigSettingsDeleter().delete(path)
        self._recentconfigs_load_defaults()

    # -- Recent remotes --
    def _recentremotes_load_defaults(self) -> None:
        with RecentConanRemotesSettingsReader() as settings:
            self._prefs_recentremotes_model.set_urls(settings.urls.resolve())

    def _recentremotes_context_menu(self, position: QtCore.QPoint) -> None:
        selected_rows = (
            self._ui.prefs_recentremotes_list.selectionModel().selectedRows()
        )
        if not selected_rows:
            return
        selected_index = selected_rows[0]
        url = self._prefs_recentremotes_model.url(selected_index.row())
        menu = QtWidgets.QMenu(self)
        menu.addActions(self._ui.prefs_recentremotes_list.actions())
        forget_config_action = menu.actions()[0]
        forget_config_action.setData(url)
        forget_config_action.triggered.connect(self._recentremotes_forget_config)
        menu.exec_(self._ui.prefs_recentremotes_list.viewport().mapToGlobal(position))

    def _recentremotes_forget_config(self) -> None:
        action = self.sender()
        url = action.data()
        RecentConanRemotesSettingsDeleter().delete(url)
        self._recentremotes_load_defaults()

    def _requires_restart(self) -> None:
        QtWidgets.QMessageBox.information(
            self,
            "Restart required",
            "Changing this preference requires restarting the application to take "
            "effect.",
        )

    def _import_presets(self) -> None:
        prefix_path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            caption="Import preferences presets",
            filter="JSON (*.json)",
        )
        if not prefix_path:
            return
        with open(prefix_path, "rt", encoding="utf-8") as preset_json_file:
            presets = json.load(preset_json_file)
        if "cruiz_presets" not in presets:
            QtWidgets.QMessageBox.critical(
                self,
                "Preference presets",
                f"{prefix_path} is an invalid preference presets file. "
                "Aborting import.",
            )
            return
        LocalCacheSettingsWriter().presets(presets["cruiz_presets"])
        self._current_load_defaults(self._ui.prefs_toolbox.currentIndex())

    def _export_presets(self) -> None:
        prefix_path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export preferences presets", "", "JSON files (*.json)"
        )
        if not prefix_path:
            return
        presets: typing.Dict[str, str] = {}
        with LocalCacheSettingsReader() as settings:
            presets.update(settings.presets())  # type: ignore
        if presets:
            with open(prefix_path, "wt", encoding="utf-8") as preset_json_file:
                json.dump({"cruiz_presets": presets}, preset_json_file, indent=4)

    def _clean_preferences(self) -> None:
        sanitise_settings(self)

    def _restore_defaults(self) -> None:
        result = QtWidgets.QMessageBox.question(
            self,
            "Factory reset",
            (
                "Performing a factory reset will clear all settings and restart cruiz. "
                "Do you want to continue?"
            ),
        )
        if result == QtWidgets.QMessageBox.StandardButton.Yes:
            factory_reset()



#!/usr/bin/env python3

"""
Conan worker message
"""

import typing


# pylint: disable=too-few-public-methods
class Message:
    """
    Base class of all messages
    """


class End(Message):
    """
    Message to indicate no more messages.
    """


class Stdout(Message):
    """
    Message for standard out data
    """

    def __init__(self, message: str) -> None:
        super().__init__()
        self._message = message

    def message(self) -> str:
        """
        Get the message
        """
        return self._message


class Stderr(Message):
    """
    Message for standard error data
    """

    def __init__(self, message: str) -> None:
        super().__init__()
        self._message = message

    def message(self) -> str:
        """
        Get the message
        """
        return self._message


class ConanLogMessage(Message):
    """
    Message for Conan logs
    """

    def __init__(self, message: str) -> None:
        super().__init__()
        self._message = message

    def message(self) -> str:
        """
        Get the message
        """
        return self._message


class Success(Message):
    """
    Message with a result payload for successful completion
    """

    def __init__(self, data: typing.Any) -> None:
        super().__init__()
        self._data = data

    def payload(self) -> typing.Any:
        """
        Get the successful result
        """
        return self._data


class Failure(Message):
    """
    Message with an exception for a failed command
    """

    def __init__(self, exception: Exception) -> None:
        super().__init__()
        self._exception = exception

    def exception(self) -> Exception:
        """
        Get the exception
        """
        return self._exception



#!/usr/bin/env python3

"""
cruiz version specification.

Add .devN to the end of the version number to indicate a development build, N commits
after the last annotated tag.
"""


def get_version() -> str:
    """
    Try to get a fixed version number from the RELEASE_VERSION module (written by
    binary distributions).
    Otherwise, fallback to accessing Git information.
    """
    try:
        from .RELEASE_VERSION import __version__  # type: ignore[import]

        return __version__
    except ImportError:
        import os
        import subprocess
        import sys

        def _describe(cwd: str) -> str:
            if (
                "GITHUB_REF_TYPE" in os.environ
                and os.environ["GITHUB_REF_TYPE"] == "tag"
            ):
                # during GitHub actions, prefer to use the ref pushed
                return os.environ["GITHUB_REF_NAME"]
            else:
                # annotated tags only
                return (
                    subprocess.check_output(["git", "describe", "--tags"], cwd=cwd)
                    .decode("utf-8")
                    .rstrip()
                )

        try:
            root_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
        except NameError:
            root_dir = os.path.dirname(sys.argv[0])
        try:
            ref_description = _describe(root_dir)
        except subprocess.CalledProcessError:
            return "0.0.0"
        ref_split = ref_description.split("-")
        # drop the v prefix from the tag
        ref_tag = ref_split[0][1:]
        if len(ref_split) == 1:
            __version__ = ref_tag
        else:
            # tag.devN where N is number of commits past the tag
            commit_count = ref_split[1]
            __version__ = f"{ref_tag}.dev{commit_count}"
        return __version__



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing
from pathlib import Path

from cruiz.interop.commandparameters import CommandParameters
from .utils import worker
from .utils.message import Message, Success, Stdout


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Delete CMakeCache.txt
    """
    with worker.Worker(queue, params):
        assert params.cwd
        cmakecache_file = params.cwd
        if params.build_folder:
            cmakecache_file /= params.build_folder
        cmakecache_file /= "CMakeCache.txt"
        try:
            # Python 3.8+
            Path(cmakecache_file).unlink(missing_ok=True)
        except TypeError:
            Path(cmakecache_file).unlink()
        queue.put(Stdout(f"Deleted {cmakecache_file}"))
        queue.put(Success(None))



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing
import subprocess

from cruiz.interop.commandparameters import CommandParameters
from .utils import worker
from .utils.message import Message, Success, Stdout, Stderr


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    # pylint: disable=too-many-branches
    """
    Run CMake build tool
    """
    assert params.cwd
    # generate build command
    if "CONAN_CMAKE_PROGRAM" in params.added_environment:
        build_cmd = [params.added_environment["CONAN_CMAKE_PROGRAM"]]
    else:
        build_cmd = ["cmake"]
    build_cmd.extend(
        [
            "--build",
            str(params.build_folder) if params.build_folder else ".",
        ]
    )
    # convert "Conan CMake" envvars into CMake envvars
    if "CONAN_MAKE_PROGRAM" in params.added_environment:
        params.added_environment["CMAKE_MAKE_PROGRAM"] = params.added_environment[
            "CONAN_MAKE_PROGRAM"
        ]
    if "CONAN_CMAKE_GENERATOR" in params.added_environment:
        params.added_environment["CMAKE_GENERATOR"] = params.added_environment[
            "CONAN_CMAKE_GENERATOR"
        ]
    if "verbose" in params.arguments:
        if params.added_environment.get("CONAN_CMAKE_GENERATOR", None) == "Ninja":
            build_cmd.append("-v")
        else:
            params.added_environment["VERBOSE"] = "1"
    if "CONAN_CPU_COUNT" in params.added_environment:
        build_cmd.append(
            f"-j{params.added_environment['CONAN_CPU_COUNT']}"
        )  # suitable for both Make and Ninja
    with worker.Worker(queue, params):
        with subprocess.Popen(
            build_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=params.cwd,
        ) as process:
            assert process.stdout
            for line in iter(process.stdout.readline, b""):
                queue.put(Stdout(line.decode("utf-8")))
            assert process.stderr
            for line in iter(process.stderr.readline, b""):
                queue.put(Stderr(line.decode("utf-8")))

            queue.put(Success(process.returncode))



#!/usr/bin/env python3

"""
Utilities for generating SVGs from Conan dependency graphs
"""

import os
import typing

import graphviz
from qtpy import QtCore, QtGui, QtSvg, QtWidgets, PYSIDE2

if PYSIDE2:
    QGraphicsSvgItem = QtSvg.QGraphicsSvgItem
    QAction = QtWidgets.QAction
else:
    # module not present in qtpy at the time of writing
    from PySide6 import QtSvgWidgets

    QGraphicsSvgItem = QtSvgWidgets.QGraphicsSvgItem
    QAction = QtGui.QAction

from cruiz.settings.managers.graphvizpreferences import GraphVizSettingsReader
from cruiz.environ import EnvironSaver
from cruiz.interop.packagenode import PackageNode
from cruiz.interop.dependencygraph import DependencyGraph


class DependenciesToDigraph:
    """
    Convert a Conan dependency graph into a GraphViz Digraph
    """

    def __init__(
        self, depgraph: DependencyGraph, rankdir: str, flipped_edges: bool = False
    ) -> None:
        graph = graphviz.Digraph(
            comment="Conan package dependency graph",
            format="svg",
            engine="dot",
        )
        graph.attr(bgcolor="lightsteelblue")
        graph.attr(rankdir=rankdir)
        graph.attr("edge", color="red")
        graph.attr("node", color="blue", style="filled", fillcolor="white")
        self._visited: typing.List[PackageNode] = []
        self._edges: typing.List[typing.Tuple[PackageNode, PackageNode]] = []
        self._visit(graph, depgraph.root, flipped_edges)
        self.digraph = graph

    def _visit(
        self, graph: graphviz.Digraph, node: PackageNode, flipped_edges: bool
    ) -> None:
        def _format_ref(ref: str) -> str:
            return ref.replace("@", "@\n")

        if node not in self._visited:
            node_name_and_id = _format_ref(node.reference)
            if node.is_runtime:
                graph.node(node_name_and_id, id=node_name_and_id)
            else:
                graph.node(node_name_and_id, fillcolor="gray", id=node_name_and_id)
            self._visited.append(node)
        for child in node.children:
            self._visit(graph, child, flipped_edges)
            if (child, node) not in self._edges:
                if flipped_edges:
                    graph.edge(
                        _format_ref(node.reference), _format_ref(child.reference)
                    )
                else:
                    graph.edge(
                        _format_ref(child.reference), _format_ref(node.reference)
                    )
                self._edges.append((child, node))


class DigraphToSVG:
    """
    Convert a Graphviz Digraph to an SVG
    """

    def __init__(self, depstodigraph: DependenciesToDigraph) -> None:
        with EnvironSaver():
            with GraphVizSettingsReader() as settings:
                graphviz_bin_dir = settings.bin_directory.resolve()
            if graphviz_bin_dir:
                os.environ["PATH"] = (
                    QtCore.QDir.toNativeSeparators(graphviz_bin_dir)
                    + os.pathsep
                    + os.environ["PATH"]
                )
            exe_path = QtCore.QStandardPaths.findExecutable("dot")
            if not exe_path:
                # if dot cannot be found, abort before trying any graphviz work
                raise FileNotFoundError(
                    "Cannot find 'dot' to generate dependency graph visualisation"
                )
            self.svg = depstodigraph.digraph.pipe()


class _SVGDialog(QtWidgets.QDialog):
    def __init__(self, renderer: QtSvg.QSvgRenderer) -> None:
        super().__init__()
        self.setWindowTitle("Package dependency graph view")
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._context_menu)
        self._renderer = renderer
        item = QGraphicsSvgItem()
        item.setSharedRenderer(renderer)
        self._scene = QtWidgets.QGraphicsScene()
        self._scene.addItem(item)
        self._view = QtWidgets.QGraphicsView(self._scene)
        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._view)

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        super().showEvent(event)
        self._view.fitInView(self._scene.sceneRect(), QtCore.Qt.KeepAspectRatio)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        self._view.fitInView(self._scene.sceneRect(), QtCore.Qt.KeepAspectRatio)
        super().resizeEvent(event)

    def _context_menu(self, position: QtCore.QPoint) -> None:
        menu = QtWidgets.QMenu(self)
        save_action = QAction("Save...", self)
        save_action.triggered.connect(self._on_save)
        menu.addAction(save_action)
        menu.exec_(self.mapToGlobal(position))

    def _on_save(self) -> None:
        new_path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save SVG", "", "SVG files (*.svg)"
        )
        if not new_path:
            return
        generator = QtSvg.QSvgGenerator()
        generator.setFileName(new_path)
        generator.setTitle("Conan package dependency graph")
        generator.setDescription("An SVG diagram of the Conan package dependency graph")
        generator.setSize(self._view.sizeHint())
        painter = QtGui.QPainter()
        if painter.begin(generator):
            self._renderer.render(painter)
            painter.end()


class SVGScene(QtWidgets.QGraphicsScene):
    """
    Wrapper around an SVG scene
    """

    def __init__(self, renderer: QtSvg.QSvgRenderer) -> None:
        super().__init__()
        self._renderer = renderer
        item = QGraphicsSvgItem()
        item.setSharedRenderer(renderer)
        self.addItem(item)

    def mouseDoubleClickEvent(self, event: QtWidgets.QGraphicsSceneMouseEvent) -> None:
        item = self.itemAt(event.scenePos(), QtGui.QTransform())
        if item and isinstance(item, QGraphicsSvgItem):
            _SVGDialog(self._renderer).exec_()
            event.setAccepted(True)
        return super().mouseDoubleClickEvent(event)



#!/usr/bin/env python3

"""
cruiz: Conan recipe user interface
"""

import os

os.environ.setdefault("QT_API", "pyside6")

from cruiz.entrypoint import main  # noqa: E402


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Exception types
"""


class RecipeInspectionError(Exception):
    """
    An error inspecting the recipe has occurred.
    """


class RecipeDoesNotExistError(Exception):
    """
    Exception class indicating a recipe does not exist.
    """


class RecipeAlreadyOpenError(Exception):
    """
    Exception class indicating a recipe is already open.
    """


class InconsistentSettingsError(Exception):
    """
    An error indicating some settings have been discovered that are inconsistent.
    """



#!/usr/bin/env python3

"""
Create an instance of the Conan API object
"""

from __future__ import annotations

import multiprocessing
import os
import pathlib
import typing

# must be performed before conan is imported
from .monkeypatch import _do_monkey_patching

_do_monkey_patching()

# pylint: disable=wrong-import-order,wrong-import-position
from conans.client import conan_api, output, runner  # noqa: E402
from conans.paths import get_conan_user_home  # noqa: E402

from .stream import QueuedStreamSix  # noqa: E402
from .message import Message, Stdout, Stderr  # noqa: E402

import cruiz.workers.utils.qtlogger  # noqa: E402
from cruiz.interop.commandparameters import CommandParameters  # noqa: E402
from cruiz.interop.commonparameters import CommonParameters  # noqa: E402


def instance(
    queue: multiprocessing.Queue[Message],
    params: typing.Union[CommandParameters, CommonParameters],
) -> conan_api.ConanAPIV1:
    """
    Get a new instance of the Conan API object
    """
    # pylint: disable=no-member
    cruiz.workers.utils.qtlogger.QtLogger().set_queue(queue)
    stdout = QueuedStreamSix(queue, Stdout)
    stderr = QueuedStreamSix(queue, Stderr)
    newoutputter = output.ConanOutput(stream=stdout, stream_err=stderr, color=True)
    if isinstance(params, CommandParameters):
        home_dir = params.added_environment.get(
            "CONAN_USER_HOME", get_conan_user_home()
        )
        local_cache_dir = os.path.join(home_dir, ".conan")
        cache = conan_api.ClientCache(local_cache_dir, stdout)
        print_commands_to_output = cache.config.print_commands_to_output
        if params.cwd:
            # TODO: this has some broken assumptions about pure paths
            if isinstance(params.cwd, pathlib.PurePosixPath):
                path = pathlib.Path(params.cwd)
                path.mkdir(parents=True, exist_ok=True)
            else:
                params.cwd.mkdir(parents=True, exist_ok=True)
            os.chdir(params.cwd)
    elif isinstance(params, CommonParameters):
        home_dir = params.added_environment.get(
            "CONAN_USER_HOME", get_conan_user_home()
        )
        local_cache_dir = os.path.join(home_dir, ".conan")
        cache = conan_api.ClientCache(local_cache_dir, stdout)
        print_commands_to_output = False

    newrunner = runner.ConanRunner(
        print_commands_to_output=print_commands_to_output, output=newoutputter
    )

    try:
        api = conan_api.ConanAPIV1(output=newoutputter, runner=newrunner)
        api.create_app()
    except TypeError:
        api = _create_old_conan_api(newoutputter, newrunner)
        api.invalidate_caches()
    return api


def _create_old_conan_api(out: typing.Any, run: typing.Any) -> typing.Any:
    # This function is mostly copied from Conan 1.17.x Factory methood for ConanAPIV1,
    # but with edits for cruiz
    # pylint: disable=logging-not-lazy,import-outside-toplevel,too-many-locals,
    # pylint: disable=unexpected-keyword-arg,no-member,no-value-for-parameter,
    # pylint: disable=too-many-function-args
    import sys
    from conans.client.userio import UserIO
    from conans.client.migrations import ClientMigrator
    from conans.client.hook_manager import HookManager
    from conans.client.cache.cache import ClientCache
    from conans.client.rest.rest_client import RestApiClient
    from conans.client.store.localdb import LocalDB
    from conans.model.version import Version
    from conans import __version__ as client_version
    from conans.util.log import configure_logger
    from conans.client.rest.conan_requester import ConanRequester
    from conans.client.rest.auth_manager import ConanApiAuthManager
    from conans.client.remote_manager import RemoteManager
    from conans.tools import set_global_instances
    from conans.client.conan_api import ConanAPIV1
    from conans.paths import get_conan_user_home
    import conans

    user_io = UserIO(out=out)

    user_home = get_conan_user_home()
    base_folder = os.path.join(user_home, ".conan")

    cache = ClientCache(base_folder, out)
    # Migration system
    migrator = ClientMigrator(cache, Version(client_version), out)
    migrator.migrate()

    if base_folder:
        sys.path.append(os.path.join(base_folder, "python"))

    config = cache.config
    # Adjust CONAN_LOGGING_LEVEL with the env readed
    conans.util.log.logger = configure_logger(config.logging_level, config.logging_file)
    conans.util.log.logger.debug("INIT: Using config '%s'", cache.conan_conf_path)

    hook_manager = HookManager(cache.hooks_path, config.hooks, user_io.out)
    # Wraps an http_requester to inject proxies, certs, etc
    requester = ConanRequester(config)
    # To handle remote connections
    put_headers = cache.read_put_headers()
    rest_api_client = RestApiClient(
        user_io.out,
        requester,
        revisions_enabled=config.revisions_enabled,
        put_headers=put_headers,
    )
    # To store user and token
    localdb = LocalDB.create(cache.localdb)
    # Wraps RestApiClient to add authentication support (same interface)
    auth_manager = ConanApiAuthManager(rest_api_client, user_io, localdb)
    # Handle remote connections
    remote_manager = RemoteManager(cache, auth_manager, user_io.out, hook_manager)

    # Adjust global tool variables
    set_global_instances(out, requester)

    # Settings preprocessor
    interactive = False

    return ConanAPIV1(
        cache,
        user_io,
        run,
        remote_manager,
        hook_manager,
        requester,
        interactive=interactive,
    )
